/* normalmode.h
 * This file belongs to Worker, a file manager for UN*X/X11.
 * Copyright (C) 2001-2013 Ralf Hoffmann.
 * You can contact me at: ralf@boomerangsworld.de
 *   or http://www.boomerangsworld.de/worker
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef NORMALMODE_H
#define NORMALMODE_H

#include "wdefines.h"
#include <aguix/request.h>
#include <aguix/message.h>
#include <aguix/refcount.hh>
#include <aguix/popupmenu.hh>
#include <map>
#include <memory>
#include "listermode.h"
#include "dirfiltersettings.hh"
#include "dirsortsettings.hh"
#include "stringmatcher_fnmatch.hh"
#include "dirbookmarkssettings.hh"
#include <string>
#include "tablookup.hh"
#include "nm_filetype_thread.hh"
#include "worker_types.h"
#include "functionproto.h"

class Lister;
class NMFieldListViewDND;
class CopyOpWin;
class NMBookmarkCB;
class NM_specialsourceExt;
class NMExtList;
class NM_CopyOp_Dir;
class NMCacheEntry;
class DeleteOpWin;
class FileEntry;
class AContainer;
class KarteiButton;
class ArgClass;
class Button;
class FieldListView;
class StringGadget;
class WCFiletype;

class NormalMode:public ListerMode
{
public:
    NormalMode( Lister * );
    virtual ~NormalMode();
    NormalMode( const NormalMode &other );
    NormalMode &operator=( const NormalMode &other );

    virtual void messageHandler( AGMessage * );
    virtual void on();
    virtual void off();
    virtual void activate();
    virtual void deactivate();
    void selhandler( FieldListView*, int row );
    virtual bool isType( const char * );
    virtual const char *getType();
    virtual int configure();
    virtual void cyclicfunc( cyclicfunc_mode_t mode );
    virtual const char* getLocaleName();
    virtual int load();
    virtual bool save( Datei * );
    virtual void relayout();
    static const char *getStaticLocaleName();
    static const char *getStaticType();
  
    int enterDir( const char *dirstr );
    int enterPath( const std::string &fullname );
    void enterDirActive();
    void up();
    void down();
    void top();
    void last();
    void pageup();
    void pagedown();
    void select();
    void selectall();
    void selectnone();
    void invertall();
    void parent();
    void filterselect( const char *filter, int mode );
    void copy( struct copyorder *copyorder );
    void deletef( struct deleteorder *delorder );
    void update( bool reset_dirsizes );
    void update( bool reset_dirsizes, bool keep_filetypes );
    void makedir();
    void renamef( struct NM_renameorder *renorder );
    void dirsize( struct NM_dirsizeorder *dsorder );
    void simdd();
    void searchentry( bool ignore_case, bool reverse_search, bool infix_search );
    void enterpath();
    void scrolllister( int dir );
    void createsymlink( struct NM_createsymlinkorder *cslorder );
    void changesymlink( struct NM_changesymlinkorder *chslorder );
    void chmodf( struct NM_chmodorder *cmorder );
    void chownf( struct NM_chownorder *coorder );
  
    void toggleHidden();
    void setShowHidden( bool );
    const char *getCurrentDir();

    int getSelFiles( std::list< NM_specialsourceExt > &list, lm_getfiles_t mode, bool unselect = false );
    int getDirSize( FileEntry *fe, loff_t &return_size, bool &cancel );

    virtual bool startdnd( DNDMsg *dm );
    virtual bool isyours( Widget *elem );

    void setSortmode( int nmode );
  
    typedef enum { NM_FILTER_INCLUDE, NM_FILTER_EXCLUDE, NM_FILTER_UNSET } nm_filter_t;
    void unsetAllFilters();
    void setFilter( const char *filter, nm_filter_t mode );
    nm_filter_t checkFilter( const char *filter ) throw( int );

    void slavehandler();
    static void freeExtList( List *extlist );

    /* deselect for extern access, check for existance of fe => slow for big dirs */
    void deselect( const FileEntry *fe );
    virtual void lvbDoubleClicked();
  
    int request( const char *title,
                 const char *text,
                 const char *buttons,
                 Requester::request_flags_t flags = Requester::REQUEST_NONE );
    int string_request( const char *title,
                        const char *lines,
                        const char *default_str,
                        const char *buttons,
                        char **return_str,
                        Requester::request_flags_t flags = Requester::REQUEST_NONE );
    int request_choose( const char *title,
                        const char *text,
                        const char *choose_text,
                        bool &choose_var,
                        const char *buttons,
                        Requester::request_flags_t flags = Requester::REQUEST_NONE );
    int request( CopyOpWin *cowin,
                 const char *title,
                 const char *text,
                 const char *buttons,
                 Requester::request_flags_t flags = Requester::REQUEST_NONE );
    int string_request( CopyOpWin *cowin,
                        const char *title,
                        const char *lines,
                        const char *default_str,
                        const char *buttons,
                        char **return_str,
                        Requester::request_flags_t flags = Requester::REQUEST_NONE );
    int request( DeleteOpWin *dowin,
                 const char *title,
                 const char *text,
                 const char *buttons,
                 Requester::request_flags_t flags = Requester::REQUEST_NONE );
    int string_request( DeleteOpWin *dowin,
                        const char *title,
                        const char *lines,
                        const char *default_str,
                        const char *buttons,
                        char **return_str,
                        Requester::request_flags_t flags = Requester::REQUEST_NONE );
    std::unique_ptr< NM_specialsourceExt > getSpecialsourceForFilename( const std::string &filename );

    void activateEntry( const std::string &name );

    bool filteredSearchActive() const;
    void deactivateFilteredSearch();
  
    void openContextMenu();
    void bookmarksChanged();
    void updateOnBookmarkChange();
    void openLabelPopUp();

    void setHighlightBookmarkPrefix( bool nv );
    bool getHighlightBookmarkPrefix() const;

    void setBookmarkFilter( DirFilterSettings::bookmark_filter_t v );
    DirFilterSettings::bookmark_filter_t getBookmarkFilter() const;

    void setSpecificBookmarkLabel( const std::string &l );
    const std::string &getSpecificBookmarkLabel() const;

    int queryLVDimensions( int &x, int &y,
                           int &w, int &h );

    void newTab();
    void closeCurrentTab();
    void closeTab( int tab );
    void switchToTab( int tab );
    void switchToTabDirection( int direction );

    void openTabPopUp( int tab );
    void enter_dir( const std::list< RefCount< ArgClass > > &args );
    void activate_entry( const std::list< RefCount< ArgClass > > &args );
    void switch_to_tab_direction( const std::list< RefCount< ArgClass > > &args );

    void setFlexibleMatchingMode( bool nv );
    bool getFlexibleMatchingMode() const;

    void recenterTopBottom();

    bool pathsChanged( const std::set< std::string > changed_paths );

    void enableDirWatch();
    void disableDirWatch();

    std::string getCurrentDirectory();

    bool deselectEntry( const std::string &filename );

    std::list< std::string > getNameOfCacheEntries();
protected:
    /*
     * NM_specialsourceInt always stores a pointer to a FE
     */
    class NM_specialsourceInt {
    public:
        NM_specialsourceInt( FileEntry *tfe );
        ~NM_specialsourceInt();
        FileEntry *entry() const
        {
            return fe;
        }
        int row;  // corresponding row or -1
        NM_CopyOp_Dir *cod; // when fe is dir, this will be the corresponding structure
    private:
        FileEntry *fe;  // element to copy
    };

    void buildListView();
    void deleteCache();
    void showcache( int );
    void showCacheAbs( int pos );
    void showCacheState();
    void startAction( FileEntry* );
  
    typedef enum { NM_COPY_OK,
                   NM_COPY_SKIP,
                   NM_COPY_CANCEL,
                   NM_COPY_NEED_DELETE } nm_copy_t;

    NMFieldListViewDND *lv;
    Button *hb[2], *parentb, *optionb;
    StringGadget *sg;
    static const char *type;
    NMCacheEntry *ce;
    List *cache;
    Requester *req;
    bool searchmodeon;
    std::string oldstr;
    std::string lastsearch;
    int sortmode;
  
    bool showfreespace;
    int updatetime;
    time_t lastfsupdate;
    std::string namestr;
    std::string freespacestr;
  
    struct timeval lastclick;
    int lastrow;
    void removeEntry( FileEntry *fe, int row );
    void fixSpecialSourceList( const std::list<NM_specialsourceInt*> *l, int deleted_row );
    void deselect( FileEntry *fe, int row );
    void startsearchmode( bool use_previous_search = false );
    void finishsearchmode();

    /**
     * jumps to the next entry matching the given string or the last search
     * string (if tstr == NULL )
     */
    bool shownextentry( const char *tstr, bool reverse_search = false, bool start_at_same = false );

    void setFilters( const std::list<NM_Filter> &filters );
    void configureFilters( bool * tshowhidden, std::list<NM_Filter> &filters, DirFilterSettings &dir_settings );
    void setLVC4Filter( FieldListView *filv, int row, NM_Filter &fi);
    int configureFilter( NM_Filter &fi );
    void rebuildView();
    void setName();
    void updateName();
    void showFreeSpace( bool force );
    int updateFreeSpaceStr( int ti );
    void setShowFreeSpace( bool v );
    void setUpdatetime( int nv );
    void setQuicksearchEnabled( bool nv );
    void setFilteredSearchEnabled( bool nv );
  
    static int maxfilestrlen, maxdirstrlen, maxbytesstrlen, maxwbytesstrlen, maxnochanges;
  
    class ft_request_list
    {
    public:
        struct ft_request_list_entry {
            std::string fullname;
            FileEntry *fe;
        };

        ft_request_list();
        ~ft_request_list();
        ft_request_list( const ft_request_list &other );
        ft_request_list &operator=( const ft_request_list &other );

        bool isFull();
        bool isEmpty();
        int put( const ft_request_list_entry &elem );
        ft_request_list_entry remove();
        ft_request_list_entry gettop();
    protected:
        std::list< ft_request_list_entry > m_elements;
        size_t m_max_elements;
    };
    void ft_request_list_clear();

    void ft_result_list_clear();

    ft_request_list m_request_list;

    void ft_list_clear();
    void ft_list_update();

    bool lasteagain;
  
    void setupLVFields();

    bool isColumnVisible( const WorkerTypes::listcol_t c );
    bool visChanged;
    int oldlvy;
  
    void changeSortModeForField( int field );
  
    int busyFlag;
    time_t lastBusyFlagUpdate;
  
    int getSelFiles( std::list<NM_specialsourceInt*> *, lm_getfiles_t, bool unselect );
    FileEntry *findFE( const FileEntry *tfe );

    bool checkForEmptyActionList( WCFiletype *ft );
    std::map<std::string,int> _field_map;
    std::map<std::string,int> _loaded_field_width;
    int getAndEraseLoadedFieldWidth( std::string str1 );
    int writeFieldWidthsToLoadMap();

    bool quicksearch_enabled;

    void makeRowActive( int row );
    void makeListViewRowActive( int row );

    DirFilterSettings _dir_filter_sets;
    DirSortSettings _dir_sort_sets;
    DirBookmarksSettings m_dir_bookmarks_sets;

    FileEntry *getFEForRow( int row );
    int _lv_ids_serial;

    StringMatcherFNMatch _entrysearch_matcher;
    void setEntrySearchString( const std::string &str, bool update_lv = true );
    std::string getEntrySearchString();
    void setEntrySearchCaseSensitive( bool nv );
  
    void updateDirFilter();

    bool _filtered_search_enabled;
    bool m_filtered_search_active;

    int buildSpecialSourceIntList( const std::list<NM_specialsourceExt*> &inlist,
                                   std::list<NM_specialsourceInt*> &outlist );
  
    void lv_pressed( int row, bool open_under_mouse = true );
    void buildContextPopUpMenu( FileEntry *entry );
    void buildLabelPopUpMenu( FileEntry *entry );
    void freePopUpMenus();
    void startLVPopUpAction( AGMessage *msg );

    time_t m_last_statebar_update;
    bool m_bookmarks_has_been_changed;
    NMBookmarkCB *m_bookmark_cb;

    std::list<PopUpMenu::PopUpEntry> buildLabelPopUpData( FileEntry *entry );
    void handleLabelPopUp( AGMessage *msg );
    void handleLabelPopUp( FileEntry *fe, const std::list<int> &entry_ids, AGMessage *msg );

    typedef struct _label_popup_table_t {
        typedef enum { REMOVE_LABEL, ADD_LABEL, ADD_CUSTOM_LABEL } label_popup_t;

        _label_popup_table_t() : type( REMOVE_LABEL ),
                                 label( "" )
        {}
        _label_popup_table_t( label_popup_t t, const std::string &l ) : type( t ),
                                                                        label( l )
        {}

        label_popup_t type;
        std::string label;
    } label_popup_table_t;

    typedef struct _tab_popup_table_t {
        typedef enum { NEW_TAB, DEL_TAB, TAB_TO_OTHER_SIDE, TAB_TO_OTHER_SIDE_AS_NEW } tab_popup_t;

        _tab_popup_table_t() : type( NEW_TAB )
        {}
        _tab_popup_table_t( tab_popup_t t ) : type( t )
        {}

        tab_popup_t type;
    } tab_popup_table_t;

    class PopUpSettings {
    public:
        PopUpSettings();
        PopUpSettings( const PopUpSettings &other );
        PopUpSettings &operator=( const PopUpSettings &rhs );

        void reset();

        int label_menu_id;
        std::map<int, label_popup_table_t> label_id_to_action;
        std::map<int, tab_popup_table_t> tab_id_to_action;
        std::unique_ptr<PopUpMenu> label_popup_menu;
        std::unique_ptr<PopUpMenu> tab_popup_menu;

        std::map<int, RefCount<ActionDescr> > popup_descr;
        std::unique_ptr<FileEntry> entry_for_popup;
        PopUpMenu *lv_popup_menu;
        int m_tab;
        std::map< int, bool > entry_is_action_submenu;
    };

    PopUpSettings m_current_popup_settings;

    AContainer *m_cont, *m_cont2, *m_tab_cont;
    KarteiButton *m_tab_b;
    Button *m_tab_new, *m_tab_close;

    void updateTabs();
    NMCacheEntry *findOldestCacheEntryNotInTab();

    struct {
        std::list<std::string> list_of_used_tabs;
        int active_tab;
    } m_off_store;

    RefCount<TextShrinker> m_shrinker;
    bool m_filtered_search_infix_search;
    TabLookup m_tab_entries;

    void setTabOption( int pos, const std::string &path );
    int addTabOption( const std::string &path );
    void removeTabOption( int pos );

    loff_t m_current_space_update_ms;
    void updateSpaceTimeout( loff_t time_ms );

    std::list<PopUpMenu::PopUpEntry> buildTabPopUpData();
    void handleTabPopUp( AGMessage *msg );

    std::string getDirectoryOfTab( int tab ) const;

    void showSearchModeCompletion( const std::string &str );
    std::string getSearchModeString();
    void resetToLastSearchString();

    std::string m_last_completion;

    std::string cleanFlexibleMatchString( const std::string &str );

    void enableDirWatch( const std::string &path );
    void setWatchMode( bool nv );

    bool m_flexible_matching_active;
    bool m_flexible_matching_enabled;

    // these variables are only used for caching
    struct {
        std::string previous_match_string;
        std::string cleaned_match_string;
    } m_cached_values;

    NM_Filetype_Thread m_ft_thread;

    int prepareContext( class ActionMessage &amsg );

    struct recenter_state {
        recenter_state() : last_state( 0 ),
                           last_yoffset( -1 ),
                           last_active_row( - 1 ) {}

        int last_state; // 0 == center, 1 == top, 2 == bottom
        int last_yoffset, last_active_row;
    } m_recenter_state;

    struct dirwatcher {
        dirwatcher() : path_watch_enabled( false ),
                       last_update( 0 ),
                       timer_enabled( false ),
                       watch_mode( true ) {}

        std::string watched_path;
        bool path_watch_enabled;

        time_t last_update;
        bool timer_enabled;

        bool watch_mode;
    } m_dirwatcher;

    struct deselect_info {
        deselect_info() : last_pos( -1 ) {}
        int last_pos;
    } m_deselect_info;
};

#endif

/* Local Variables: */
/* mode:c++ */
/* End: */
