/* normalmode_delete.cc
 * This file belongs to Worker, a file manager for UN*X/X11.
 * Copyright (C) 2011-2013 Ralf Hoffmann.
 * You can contact me at: ralf@boomerangsworld.de
 *   or http://www.boomerangsworld.de/worker
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "normalmode.h"
#include "worker_locale.h"
#include "deleteop.h"
#include "nmcacheentry.hh"
#include "nmcopyopdir.hh"
#include "nmfieldlistviewdnd.hh"
#include "fileentry.hh"
#include "deletecore.hh"
#include "deleteorder.hh"

void NormalMode::deletef( struct deleteorder *delorder )
{
    NM_specialsourceInt *ss1;
    //  bool found;
    bool cancel=false;
    DeleteOpWin *dowin;
    char *buttonstr,*textstr;
    int erg;
    int tnrfiles,tnrdirs;
    std::list<NM_specialsourceInt*>::iterator iti1;
  
    if(delorder==NULL) return;
    if ( getCurrentDir() == NULL ) {
        // no current dir => no files to delete
        return;
    }

    if ( ce->dirOpened() == false ) return;

    tnrfiles = ce->getNrOfFiles( 1 );
    tnrdirs = ce->getNrOfDirs( 1 );

    if ( ( delorder->source == delorder->NM_ALLENTRIES ) &&
         ( ( tnrfiles + tnrdirs ) < 1 ) ) {
        if ( ce->getActiveFE() != NULL ) {
            if ( strcmp( ce->getActiveFE()->name, ".." ) != 0 ) {
                if ( ce->getActiveFE()->isDir() == true ) tnrdirs++;
                else tnrfiles++;
            }
        }
    }
  
    if ( ( tnrfiles + tnrdirs ) < 1 )
        return;

    finishsearchmode();
  
    buttonstr=(char*)_allocsafe(strlen(catalog.getLocale(11))+1+
                                strlen(catalog.getLocale(8))+1);
    sprintf(buttonstr,"%s|%s",catalog.getLocale(11),
            catalog.getLocale(8));
    textstr = (char*)_allocsafe( strlen( catalog.getLocale( 145 ) ) + A_BYTESFORNUMBER( tnrfiles ) +
                                 A_BYTESFORNUMBER( tnrdirs ) + 1 );
    sprintf(textstr,catalog.getLocale(145),tnrfiles,tnrdirs);
    erg = request( catalog.getLocale( 123 ), textstr, buttonstr );
    _freesafe(buttonstr);
    _freesafe(textstr);
    if(erg!=0) return;

    disableDirWatch();
  
    // clear the reclist so the slace doesn't block us because he is still reading
    // from disk
    ft_request_list_clear();
  
    dowin=delorder->dowin;
    if(dowin!=NULL) {
        dowin->open();
        dowin->setmessage(catalog.getLocale(140));
        if ( dowin->redraw() != 0 ) cancel = true;
    }

    DeleteCore *dc = new DeleteCore( delorder );
    int row_adjustment = 0;

    if ( cancel ) dc->setCancel( cancel );

    {    
        std::list<NM_specialsourceInt*> dellist;

        switch(delorder->source) {
            case deleteorder::NM_SPECIAL:
                break;
            case deleteorder::NM_ONLYACTIVE:
                break;
            case deleteorder::NM_ONLYSELECTED:
                getSelFiles( &dellist, LM_GETFILES_ONLYSELECT, false );
                break;
            default:  // all selected entries
                getSelFiles( &dellist, LM_GETFILES_SELORACT, false );
                break;
        }
        for ( auto &it : dellist ) {
            dc->registerFEToDelete( *it->entry(), it->row );
        }

        for ( iti1 = dellist.begin(); iti1 != dellist.end(); iti1++ ) {
            ss1 = *iti1;
            if(ss1->cod!=NULL) delete ss1->cod;
            delete ss1;
        }
    }

    if (dc->buildDeleteDatabase() != 0) dc->setCancel( true );

    if (dowin!=NULL) dowin->redraw();

    dc->setPreDeleteCallback( [this,
                               &row_adjustment]( const FileEntry &fe, int row, const NM_CopyOp_Dir *cod )
                              {
                                  int actual_row = row - row_adjustment;
                                  if ( actual_row >= 0 ) {
                                      lv->setVisMark( actual_row, true );
                                      lv->showRow( actual_row, false );
                                      lv->redraw();
                                  }
                              } );

    dc->setPostDeleteCallback( [this,
                                &delorder,
                                &dc,
                                &row_adjustment]( const FileEntry &fe, int row,
                                                  DeleteCore::nm_delete_t res, const NM_CopyOp_Dir *cod )
                               {
                                   int actual_row = row - row_adjustment;

                                   FileEntry *real_fe = getFEForRow( actual_row );

                                   if ( real_fe && real_fe->equals( fe ) ) {

                                       lv->setVisMark( actual_row, false );

                                       if ( res == DeleteCore::NM_DELETE_OK &&
                                            ( cod == NULL || cod->error_counter == 0 ) ) {
                                           // success so remove entry
                                           removeEntry( real_fe, actual_row );

                                           row_adjustment++;
                                       } else if ( res == DeleteCore::NM_DELETE_SKIP ) {
                                       } else if ( res == DeleteCore::NM_DELETE_CANCEL ) {
                                           dc->setCancel( true );
                                       }
                                   }
                               } );

    if ( ! dc->getCancel() ) {
        dc->executeDelete();
    }

    delete dc;
    dc = NULL;

    if(dowin!=NULL) dowin->close();
  
    ce->restartcheck();
    rebuildView();
    lv->showActive();
    aguix->Flush();

    enableDirWatch();
}
